<?php

use Drupal\drupalmoduleupgrader\ModuleContext;
use Symfony\Component\Yaml\Yaml;

/**
 * Implements hook_drush_command().
 */
function drupalmoduleupgrader_drush_command() {
  $items = [];

  $items['dmu-list'] = [
    'description' => "Lists available analysis/conversion routines.",
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_ROOT,
  ];
  $items['dmu-analyze'] = [
    'description' => "Analyzes a Drupal 7 module and reports the changes needed to port it to Drupal 8.",
    'arguments' => [
      'module' => 'The machine name of a Drupal 7 module installed in the modules directory.',
    ],
    'required-arguments' => TRUE,
    'options' => [
      'only' => [
        'description' => 'A comma-separated list of analysis routines to run, excluding all others.',
        'example-value' => 'HookMenu,VariableAPI,BlockInfo',
        'value' => 'optional',
      ],
      'skip' => [
        'description' => 'A comma-separated list of analysis routines to skip.',
        'example-value' => 'HookInit,HookExit',
        'value' => 'optional',
      ],
    ],
    'examples' => [
      'drush dmu-analyze pants' => 'Analyze what needs to be changed in order to port the pants module.',
    ],
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_ROOT,
  ];
  $items['dmu-upgrade'] = [
    'description' => "Upgrades a Drupal 7 module to Drupal 8.",
    'arguments' => [
      'module' => 'The machine name of a Drupal 7 module installed in the modules directory.',
    ],
    'required-arguments' => TRUE,
    'options' => [
      'only' => [
        'description' => 'A comma-separated list of conversion routines to run, excluding all others.',
        'example-value' => 'HookMenu,VariableAPI,BlockInfo',
        'value' => 'optional',
      ],
      'skip' => [
        'description' => 'A comma-separated list of conversion routines to skip.',
        'example-value' => 'HookInit,HookExit',
        'value' => 'optional',
      ],
    ],
    'examples' => [
      'drush dmu-upgrade pants' => 'Upgrade whatever can be automatically upgraded in the pants module.',
    ],
    'bootstrap' => DRUSH_BOOTSTRAP_DRUPAL_ROOT,
  ];

  return $items;
}

/**
 * Creates a new ModuleContext for a D7 module.
 *
 * @param string $module
 *  The machine name of the module.
 *
 * @return ModuleContext
 */
function _dmu_module_context($module) {
  require_once 'modules/drupalmoduleupgrader/vendor/autoload.php';
  return new ModuleContext($module);
}

/**
 * Loads all module-wide plugins.
 *
 * @param ModuleContext $context
 *  The module we're working with.
 *
 * @return array
 */
function _dmu_plugin_list() {
  // Get the conversion priority list. Conversions should generally be written
  // in a way that allows them to run in any order, but certain ones, like Grep
  // and FunctionReplacement, need to run first. The priority array (defined in
  // config/install/drupalmoduleupgrader.settings.yml) lists the ones that should
  // run before any others, and the order in which they should be executed.
  $priority = \Drupal::config('drupalmoduleupgrader.settings')->get('priority');

  // Instantiate the plugin manager, get all available plugin IDs, and sort them
  // by the contents of the priority list.
  $manager = \Drupal::service('plugin.manager.drupalmoduleupgrader.module');
  $plugin_IDs = array_keys($manager->getDefinitions());
  $plugin_IDs = array_replace(array_flip($priority), array_flip($plugin_IDs));
  $plugin_IDs = array_keys($plugin_IDs);

  // If the user has specified a list of conversions to include or exclude,
  // filter the list of plugin IDs by that list.
  if ($only = drush_get_option('only', FALSE)) {
    $plugin_IDs = array_intersect($plugin_IDs, explode(',', $only));
  }
  elseif ($skip = drush_get_option('skip', FALSE)) {
    $plugin_IDs = array_diff($plugin_IDs, explode(',', $skip));
  }

  return $plugin_IDs;
}

/**
 * Lists all the available module-wide plugins.
 */
function drush_drupalmoduleupgrader_dmu_list() {
  $manager = \Drupal::service('plugin.manager.drupalmoduleupgrader.module');

  $list = [];
  foreach ($manager->getDefinitions() as $id => $definition) {
    $list[$id] = $definition['description'];
  }
  drush_print_table(drush_key_value_to_array_table($list));
}

/**
 * Analyzes what needs changing in a module to port it to Drupal 8.
 *
 * @param string $module
 *  The machine name of the module to analyze.
 */
function drush_drupalmoduleupgrader_dmu_analyze($module) {
  $module = _dmu_module_context($module);
  $plugin_manager = \Drupal::service('plugin.manager.drupalmoduleupgrader.module');

  foreach (_dmu_plugin_list() as $plugin_id) {
    $plugin = $plugin_manager->createInstance($plugin_id, [ 'module' => $module ]);

    if ($plugin->isExecutable()) {
      $plugin->analyze();
    }
  }

  $module->generateReport();
  drush_log('Generated a report in ' . $module->getPath('upgrade-info.html'), 'success');
}

/**
 * Tries to automatically convert a Drupal 7 module to Drupal 8.
 *
 * @param string $module
 *  The module to upgrade.
 */
function drush_drupalmoduleupgrader_dmu_upgrade($module) {
  $module = _dmu_module_context($module);
  $plugin_manager = \Drupal::service('plugin.manager.drupalmoduleupgrader.module');

  foreach (_dmu_plugin_list() as $plugin_id) {
    $plugin = $plugin_manager->createInstance($plugin_id, [ 'module' => $module ]);

    if ($plugin->isExecutable()) {
      try {
        $plugin->convert();
      }
      catch (\Exception $e) {
        drush_log('[' . $plugin->getPluginId() . '] ' . $e->getMessage(), 'error');
      }
    }
  }

  $module->save();
}
