<?php

namespace Drupal\drupalmoduleupgrader\Converter;

use Drupal\Core\Plugin\PluginBase;
use Drupal\drupalmoduleupgrader\ModuleContext;
use Pharborist\Parser;
use Pharborist\Filter;

/**
 * Base class for all converters.
 */
abstract class ConverterBase extends PluginBase implements ConverterInterface {

  /**
   * The module being worked on.
   *
   * @var \Drupal\drupalmoduleupgrader\ModuleContext
   */
  protected $module;

  /**
   * If this converter uses delegates, this is the delegate manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  protected $delegator;

  /**
   * Identifiers for operations which have already been performed, in order
   * to prevent duplicate operations.
   *
   * @var string[]
   */
  protected static $processed = [];

  /**
   * @var \Drupal\Core\Database\Connection
   */
  protected $db;

  /**
   * Constructs a ConverterBase object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    // The issue_type is used by the analyze() method to group violations in
    // the generated report. If it's not explicitly given, use the plugin ID.
    if (empty($plugin_definition['issue_type'])) {
      $plugin_definition['issue_type'] = $plugin_id;
    }
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    // All converters need a ModuleContext. If we didn't get one, throw a tantrum.
    // @todo This should be an argument, not a configuration option.
    if (isset($configuration['module']) && $configuration['module'] instanceof ModuleContext) {
      $this->module = $configuration['module'];
    }
    else {
      throw new \UnexpectedValueException('Cannot instantiate converter without a ModuleContext.');
    }

    // If this converter uses delegates, get the delegates' plugin manager.
    if (isset($plugin_definition['delegator'])) {
      $this->delegator = \Drupal::service($plugin_definition['delegator']);
    }

    // Get a handle to the database (later on this, will allow for container
    // injection and unit testing).
    $this->db = \Drupal::database();
  }

  /**
   * {@inheritdoc}
   */
  public function isExecutable() {
    return TRUE;
  }

  /**
   * {@inheritdoc}
   */
  public function analyze() {
    $definition = $this->pluginDefinition;

    if (isset($definition['message'])) {
      $this->module->setIssue($this, $definition['issue_type'], $definition['message'], NULL, $definition['level'], $definition['documentation']);
    }
  }

  /**
   * @param string $id
   *  The delegate plugin ID.
   * @param array $configuration
   *  Additional configuration to be handed to the delegate. The current module
   *  will be automatically added.
   *
   * @return object
   *  The delegate plugin instance.
   */
  protected function getDelegate($id, array $configuration = []) {
    $configuration['module'] = $this->module;
    return $this->delegator->createInstance($id, $configuration);
  }

  /**
   * Parses a generated class into a syntax tree.
   *
   * @param string|array $class
   *  The class to parse, either as a string of PHP code or a renderable array.
   *
   * @return \Pharborist\ClassNode
   */
  protected function parse($class) {
    if (is_array($class)) {
      $class = drupal_render($class);
    }
    return Parser::parseSnippet($class)->find(Filter::isInstanceOf('Pharborist\ClassNode'))[0];
  }

  /**
   * Gets or sets the status of this plugin (i.e., if it has already been
   * processed).
   *
   * @param bool|null $status
   *  If boolean, marks the plugin as processed or unprocessed.
   *
   * @return bool|null
   *  If $status is passed, no return value. Otherwise, returns the status of
   *  the plugin.
   */
  protected function isProcessed($status = NULL) {
    if (isset($status)) {
      if ($status) {
        static::$processed[] = $this->getHash();
      }
      else {
        $index = array_search($this->getHash(), static::$processed);
        if (is_integer($index)) {
          unset(static::$processed[$index]);
        }
      }
    }
    else {
      return in_array($this->getHash(), static::$processed);
    }
  }

  /**
   * Returns an arbitrary hash to identify this plugin instance/configuration.
   *
   * @return string
   */
  protected function getHash() {
    return $this->pluginId;
  }

}
