<?php

namespace Drupal\drupalmoduleupgrader\Converter\Routing;

use Drupal\Component\Plugin\Exception\PluginNotFoundException;
use Drupal\drupalmoduleupgrader\Converter\ConverterBase;
use Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\ParameterMap;
use Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\RouteWrapper as Drupal7Route;
use Drupal\drupalmoduleupgrader\Converter\Routing\Drupal8\RouteWrapper as Drupal8Route;
use Drupal\drupalmoduleupgrader\Utility\StringTransformTrait;
use Pharborist\ClassMethodCallNode;
use Pharborist\Filter;
use Pharborist\FunctionDeclarationNode;
use Pharborist\ReturnStatementNode;
use Symfony\Component\Routing\Route as CoreRoute;

/**
 * Base class for route converters, which are usually delegates of
 * \Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide\HookMenu.
 */
abstract class RouteConverterBase extends ConverterBase implements RouteConverterInterface {

  use StringTransformTrait;

  /**
   * The controller into which callback functions are transferred.
   *
   * @var \Pharborist\ClassNode
   */
  protected $controller;

  /**
   * @var \Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\Route
   */
  protected $route;

  /**
   * The generated route name.
   *
   * @var string
   */
  protected $name;

  /**
   * Constructs a RouteConverterBase object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    // Use the default controller generated by HookMenu. Subclasses can
    // swap in their own controller, if need be (i.e., FormRoute).
    $this->controller = $this->module->classes['Controller\DefaultController'];

    if ($configuration['route'] instanceof Drupal7Route) {
      $this->route = $configuration['route'];
      $this->arguments = $this->route['page arguments'];
    }
    else {
      throw new \InvalidArgumentException("Route converters cannot be instantiated without a route definition.");
    }

    $title_callback = $this->route['title callback'];
    if ($this->module->defines('function', $title_callback)) {
      $this->route['title callback'] = clone $this->module->getFunction($title_callback);
      $this->route['title arguments'] = new ParameterMap($this->route->getPath(), $this->route['title callback']->getParameters(), $this->route['title arguments']);
    }

    $access_callback = $this->route['access callback'];
    if ($this->module->defines('function', $access_callback)) {
      $this->route['access callback'] = clone $this->module->getFunction($access_callback);
      $this->route['access arguments'] = new ParameterMap($this->route->getPath(), $this->route['access callback']->getParameters(), $this->route['access arguments']);
    }

    $page_callback = $this->route['page callback'];
    if ($this->module->defines('function', $page_callback)) {
      // Generate the route name from the page callback and arguments, prefixed with the
      // module's machine name.
      $module_name = $this->module->getMachineName();
      $this->name = $module_name . '.' . $this->unPrefix($page_callback, $module_name);
      $arguments = array_filter($this->route['page arguments'], 'is_string');
      if ($arguments) {
        $this->name .= '_' . implode('_', $arguments);
      }

      $this->route['page callback'] = clone $this->module->getFunction($page_callback);
      $this->route['page arguments'] = new ParameterMap($this->route->getPath(), $this->route['page callback']->getParameters(), $this->route['page arguments']);

    }
  }

  /**
   * {@inheritdoc}
   */
  public function getName() {
    return $this->name;
  }

  /**
   * {@inheritdoc}
   */
  public function buildRoute() {
    $route = new CoreRoute('');

    $map = new ParameterMap(clone $this->route->getPath(), []);

    if ($this->route['title arguments'] instanceof ParameterMap) {
      $map->merge($this->route['title arguments']);
    }
    if ($this->route['access arguments'] instanceof ParameterMap) {
      $map->merge($this->route['access arguments']);
    }
    if ($this->route['page arguments'] instanceof ParameterMap) {
      $map->merge($this->route['page arguments']);
    }
    $map->apply($route);

    if ($this->route->containsKey('title')) {
      $route->setDefault('_title', $this->route['title']);
    }
    elseif ($this->route['title callback'] instanceof FunctionDeclarationNode) {
      $route->setDefault('_title_callback', $this->getQualifiedMethodName($this->route['title callback']->getName()));
    }

    if ($this->route->isAbsoluteAccess()) {
      $route->setRequirement('_access', $this->route['access callback'] ? 'true' : 'false');
    }
    elseif ($this->route->isPermissionBased()) {
      $route->setRequirement('_permission', $this->route['access arguments'][0]);
    }
    elseif ($this->route['access callback'] instanceof FunctionDeclarationNode) {
      $route->setRequirement('_custom_access', $this->getQualifiedMethodName($this->route['access callback']->getName()));
    }

    return new Drupal8Route($this->getName(), $route);
  }

  /**
   * {@inheritdoc}
   */
  public function convert() {
    if ($this->route['title callback'] instanceof FunctionDeclarationNode) {
      $this->transferToController($this->route['title callback']);
    }
    if ($this->route['access callback'] instanceof FunctionDeclarationNode) {
      $returns = $this->route['access callback']->find(Filter::isInstanceOf('\Pharborist\ReturnStatementNode'));
      foreach ($returns as $ret) {
        $call = ClassMethodCallNode::create('\Drupal\Core\Access\AccessResult', 'allowedIf')->appendArgument($ret->getExpression());
        $ret->replaceWith(ReturnStatementNode::create($call));
      }

      $this->transferToController($this->route['access callback']);
    }
  }

  /**
   * Transfers a function into the controller, if it hasn't already been.
   *
   * @param \Pharborist\FunctionDeclarationNode $function
   *  The function to transfer.
   *
   * @return mixed
   *  The fully qualified method name.
   */
  protected function transferToController(FunctionDeclarationNode $function) {
    $function_name = $function->getName()->__toString();

    if (! in_array($function_name, static::$processed)) {
      static::$processed[] = $function_name;

      $function
        ->cloneAsMethodOf($this->controller)
        ->setName($this->module->getMethodName($function_name));
    }

    return $this->getQualifiedMethodName($function_name);
  }

  /**
   * Converts a string to a fully qualified method name in the controller.
   *
   * @param string $string
   *  The string to convert (usually a function name).
   *
   * @return string
   */
  protected function getQualifiedMethodName($string) {
    return $this->module->getNS() . '\\Controller\\DefaultController::' . $this->module->getMethodName($string);
  }

}
