<?php

/**
 * @file
 * Contains Drupal\drupalmoduleupgrader\ModuleContext.
 */

namespace Drupal\drupalmoduleupgrader;

use Drupal\drupalmoduleupgrader\Converter\ConverterInterface;
use Drupal\drupalmoduleupgrader\Utility\StringTransformTrait;
use Pharborist\ClassNode;
use Pharborist\DocCommentNode;
use Pharborist\FunctionCallNode;
use Pharborist\FunctionDeclarationNode;
use Pharborist\Filter;
use Pharborist\Parser;
use Pharborist\SourcePosition;
use Pharborist\WhitespaceNode;
use Symfony\Component\Yaml\Yaml;

/**
 * Contains the state of the module upgrade, along with useful methods for conversion
 * routines to use.
 */
class ModuleContext {

  use StringTransformTrait;

  /**
   * Syntax trees for currently open module files, keyed by path.
   *
   * @var \Pharborist\TopNode[]
   */
  private $files = [];

  /**
   * Classes defined by the module, including tests.
   *
   * @var \Pharborist\ClassNode[]
   *
   * @deprecated
   *  Should be deleted when addClass() is refactored to, like, not use it.
   */
  public $classes = [];

  /**
   * The module's machine name.
   *
   * @var string
   */
  protected $machineName;

  /**
   * The module's base path.
   *
   * @var string
   */
  protected $basePath;

  /**
   * Services provided by the module in MODULE.services.yml.
   *
   * @var array
   */
  public $services = [];

  /**
   * Form controllers, keyed by form ID and written out in src/Form.
   *
   * @var \Pharborist\ClassNode[]
   */
  public $forms = [];

  /**
   * Stores information about issues identified by plugins during analysis.
   *
   * @var array
   */
  private $issues = [];

  /**
   * @var \Drupal\Core\Database\Connection
   */
  private $db;

  /**
   * Constructs a ModuleContext object.
   *
   * @param string $machine_name
   *  The machine name of the module.
   *
   * @todo
   *  Start with the base path, and derive the machine name.
   */
  public function __construct($machine_name) {
    $this->machineName = $machine_name;

    $dir = 'modules/' . $machine_name;
    if (is_dir($dir)) {
      $this->basePath = $dir;
    }
    else {
      throw new \RuntimeException("Cannot find module directory {$dir}.");
    }

    $this->db = \Drupal::database();
    $this->index();
  }

  /**
   * Saves the modified module, writing out all configuration and classes.
   *
   * @param boolean $reindex
   *  If TRUE, rebuild the index after saving everything.
   */
  public function save($reindex = FALSE) {
    if ($this->services) {
      $this->writeInfo('services', [ 'services' => $this->services ]);
    }

    // If any form controllers have been created, write them out too.
    foreach ($this->forms as $form_controller) {
      $this->write('src/Form/' . $form_controller->getName() . '.php', $form_controller->parents()[0]);
    }

    foreach ($this->files as $path => $doc) {
      file_put_contents($path, $doc);
    }
    $this->releaseFiles();

    if ($reindex) {
      $this->index();
    }
  }

  /**
   * Indexes the module, parsing its various files into syntax trees and putting info
   * about its various code objects (functions, classes, etc.) into the database.
   */
  public function index() {
    // Clear the index for this module.
    $this->db
      ->delete('dmu_index')
      ->condition('module', $this->getMachineName())
      ->execute();

    // @todo This presumes that there are no sub-modules in the module directory.
    // It needs to be smart enough to tell 'em apart (i.e., by the presence of
    // other info files).
    $files = file_scan_directory($this->getPath(), '/.(module|install|inc|php|test)$/');
    foreach ($files as $file) {
      $this->indexObject('file', $file->filename, $file);

      // Parse the file so we can find code bits...tasty, tasty code bits
      // ...to index.
      $tree = Parser::parseFile($file->uri);

      $tree
        ->find(Filter::isInstanceOf('\Pharborist\FunctionCallNode'))
        ->each(function(FunctionCallNode $call) use ($file) {
          $this->indexObject('function_call', $call->getName(), $file);
        });

      $tree
        ->find(Filter::isInstanceOf('\Pharborist\ClassNode'))
        ->each(function(ClassNode $class) use ($file) {
          $this->indexObject('class', $class->getName(), $file);
        });

      // Scan for supported hooks. Eventually this should be dynamic, looking
      // for plugins which implement hooks (i.e., a method in the ModuleWide
      // plugin manager or whatevs).
      $hooks = [
        'block_info', 'element_info', 'exit', 'form_alter', 'forms', 'help',
        'init', 'library_info', 'menu', 'menu_alter', 'permission', 'uninstall',
        'url_outbound_alter'];
      $tree
        ->find(Filter::isInstanceOf('\Pharborist\FunctionDeclarationNode'))
        ->each(function(FunctionDeclarationNode $function) use ($file) {
          $this->indexObject('function', $function->getName(), $file);
        })
        ->filter(function(FunctionDeclarationNode $function) use ($hooks) {
          $function_name = $function->getName()->__toString();
          foreach ($hooks as $hook) {
            if ($function_name == $this->getMachineName() . '_' . $hook) {
              return TRUE;
            }
          }
          return FALSE;
        })
        ->each(function(FunctionDeclarationNode $function) use ($file) {
          $module = $this->getMachineName();
          $this->indexObject('hook', subStr((string) $function->getName(), strLen($module) + 1), $file);
        });
    }
  }

  /**
   * Puts an object object in the index.
   *
   * @param string $type
   *  The object type. Can be any arbitrary string.
   * @param string $name
   *  The name of the object. (Can also be an object that can be casted to
   *  a string.)
   * @param \StdClass $file
   *  The file in which the object resides.
   */
  private function indexObject($type, $name, \StdClass $file = NULL) {
    $this->db
      ->insert('dmu_index')
      ->fields([
        'module' => $this->getMachineName(),
        'type' => $type,
        'name' => (string) $name,
        'file' => $file ? $file->uri : $this->getPath('.module'),
      ])
      ->execute();
  }

  /**
   * Returns the module's machine name.
   *
   * @return string
   */
  public function getMachineName() {
    return $this->machineName;
  }

  /**
   * Returns the path to the module, or one of its files/directories.
   *
   * @param string|null $file
   *  The file whose path to get, or NULL to get the base path. If the file begins
   *  with a period, the module's machine name will be prepended to it automatically.
   *  So, to get the path to the install file, you could call $this->getPath('.install')
   *
   * @return string
   */
  public function getPath($file = NULL) {
    if (isset($file)) {
      if ($file{0} == '.') {
        $file = $this->getMachineName() . $file;
      }
      return $this->basePath . '/' . $file;
    }
    else {
      return $this->basePath;
    }
  }

  /**
   * Returns this module's root namespace: Drupal\MODULE (no leading backslash).
   *
   * @return string
   */
  public function getNS() {
    return 'Drupal\\' . $this->getMachineName();
  }

  /**
   * Converts a string to a method name by converting it to camelCase and removing the
   * module prefix, if it's there.
   *
   * @param string $string
   *  The string to convert (usually a function name).
   *
   * @return string
   */
  public function getMethodName($string) {
    return $this->toCamelCase($this->unPrefix($string, $this->getMachineName()));
  }

  /**
   * Returns the module's implementation of the given hook, if it exists.
   *
   * @param string $hook
   *  The hook to return, without the hook_ prefix.
   *
   * @return \Pharborist\FunctionDeclarationNode|null
   */
  public function getHook($hook) {
    return $this->getFunction($this->getMachineName() . '_' . $hook);
  }

  /**
   * Convenience method to execute this module's implementation of a hook.
   *
   * @param string $hook
   *  The hook to execute, without the hook_ prefix.
   *
   * @return mixed
   */
  public function executeHook($hook) {
    return $this->executeFunction($this->getHook($hook));
  }

  /**
   * Returns if this module defines a particular object in the index.
   *
   * @param string $type
   *  The object type.
   * @param string $name
   *  The object name. (Can also be an object that casts to string.)
   *
   * @return boolean
   */
  public function defines($type, $name) {
    return (boolean) $this->db
      ->select('dmu_index', 'i')
      ->condition('module', $this->getMachineName())
      ->condition('type', $type)
      ->condition('name', $name)
      ->fields('i')
      ->countQuery()
      ->execute()
      ->fetchField();
  }

  public function getFileContaining($type, $name) {
    $file = $this->db
      ->select('dmu_index', 'i')
      ->condition('module', $this->getMachineName())
      ->condition('type', $type)
      ->condition('name', $name)
      ->fields('i', ['file'])
      ->execute()
      ->fetchField();

    return $this->getFile($file);
  }

  /**
   * Gets the distinct names of all indexed objects of a particular type.
   *
   * @param string $type
   *  The object type.
   *
   * @return string[]
   */
  public function getObjectNames($type) {
    return $this->db
      ->select('dmu_index', 'i')
      ->distinct()
      ->fields('i', ['name'])
      ->condition('module', $this->getMachineName())
      ->condition('type', $type)
      ->execute()
      ->fetchCol();
  }

  /**
   * Returns a Pharborist node for a function defined by this module.
   *
   * @param string $function
   *  The function name.
   *
   * @return FunctionDeclarationNode
   *
   * @throws \Drupal\drupalmoduleupgrader\UndefinedFunctionException
   */
  public function getFunction($function) {
    return $this->getFileContaining('function', $function)->children(Filter::isFunction($function))[0];
  }

  /**
   * Executes an arbitrary function.
   *
   * @param \Pharborist\FunctionDeclarationNode $function
   *  The function to execute.
   * @param mixed ...
   *  Additional arguments to pass to the function.
   *
   * @return mixed
   */
  public function executeFunction(FunctionDeclarationNode $function) {
    $function_name = $function->getName()->__toString();

    if (! function_exists($function_name)) {
      // eval() is hateful, but I can't think of a better way to import the function
      // without potentially causing fatal errors due to poorly written code (i.e.,
      // class definitions that extend non-existent classes...globally-scoped function
      // calls to non-existent functions...other ghoulish things...)
      eval($function->__toString());
    }

    $arguments = array_slice(func_get_args(), 1);
    return call_user_func_array($function_name, $arguments);
  }

  /**
   * Tests if this module has a controller for a specific form.
   *
   * @param string $form_id
   *  The form ID (not the name of the controller class).
   *
   * @return bool
   */
  public function definesForm($form_id) {
    return isset($this->forms[$form_id]) && $this->forms[$form_id] instanceof ClassNode;
  }

  /**
   * Returns a form controller class (as a Pharborist node) for a specific
   * form ID.
   *
   * @param string $form_id
   *  The form ID.
   *
   * @return \Pharborist\ClassNode
   *
   * @throws \Drupal\drupalmoduleupgrader\UndefinedFormException
   */
  public function getForm($form_id) {
    if ($this->definesForm($form_id)) {
      return $this->forms[$form_id];
    }
    else {
      throw new UndefinedFormException("Undefined form {$form_id}");
    }
  }

  /**
   * Adds a service definition.
   *
   * @param string $key
   *  The service's machine name.
   * @param array $definition
   *  The service definition.
   */
  public function addService($key, array $definition) {
    $this->services[$key] = $definition;
  }

  /**
   * Write a file into the module's directory.
   *
   * @param string $destination
   *  The destination file, relative to the module root. (To automatically prefix
   *  the module's machine name, pass the file extension only.)
   * @param string $contents
   */
  public function write($destination, $contents) {
    $destination = $this->getPath($destination);

    $directory = dirName($destination);
    if (file_prepare_directory($directory, FILE_CREATE_DIRECTORY | FILE_MODIFY_PERMISSIONS)) {
      file_put_contents($destination, $contents);
    }
    else {
      throw new \RuntimeException('Cannot write ' . $destination);
    }
  }

  /**
   * Writes info to a YAML file.
   *
   * @param string $name
   *  The name of the YAML file. It will actually be written as MODULE.$name.yml.
   * @param array $data
   *  The data to write.
   */
  public function writeInfo($name, array $data) {
    $this->write('.' . $name . '.yml', Yaml::dump($data, 4, 2));
  }

  /**
   * Generates an empty hook implementation.
   *
   * @param string $hook
   *  The hook to implement, without the hook_ prefix.
   *
   * @return \Pharborist\FunctionDeclarationNode
   *  The generated hook implementation, or the existing one.
   */
  public function implementHook($hook) {
    if ($this->defines('hook', $hook)) {
      return $this->getHook($hook);
    }
    else {
      $function = FunctionDeclarationNode::create($this->getMachineName() . '_' . $hook);
      $function->setDocComment(DocCommentNode::create('Implements hook_' . $hook . '().'));
      $function->appendTo($this->fileManager[$this->getMachineName() . '.module']);
      WhitespaceNode::create("\n")->insertBefore($function);
      WhitespaceNode::create("\n")->insertAfter($function);

      return $function;
    }
  }

  /**
   * Marks a class as belonging to this module.
   *
   * @param \Pharborist\ClassNode $class
   *  The class to add. It will keyed by its path relative to this module's
   *  root namespace.
   */
  public function addClass(ClassNode $class) {
    $key = subStr($class->getName()->getAbsolutePath(), 1);
    $key = substr($key, strLen($this->getNS()));
    $key = ltrim($key, '\\');
    $this->classes[$key] = $class;
  }

  /**
   * Writes out a class to disk. The path of the class file is determined by the
   * class' fully qualified name.
   *
   * @param \Pharborist\ClassNode $class
   */
  public function writeClass(ClassNode $class) {
    // We'll use the class' fully qualified path to determine where it needs
    // to be written.
    $path = str_replace([ $this->getNS(), '\\' ], [ 'src/', '/' ], $class->getName()->getAbsolutePath());
    $this->write($path . '.php', $class->parents()[0]);
  }

  /**
   * Flags an issue, which may or may not contain any violations. An issue must exist
   * -- identified by $issue_type -- before any violations can be logged against it.
   * If the given $issue_type already exists, the existing issue will not be overwritten.
   *
   * @param \Drupal\drupalmoduleupgrader\Converter\ConverterInterface $converter
   *  The plugin which is flagging the issue.
   * @param string $issue_type
   *  The issue type. Should be a unique, constant string.
   * @param string $title
   *  The issue's title (HTML allowed).
   * @param string $summary
   *  A detailed summary of the issue (HTML allowed).
   * @param string $level
   *  The error level. This is only used as the CSS class when rendering the report.
   * @param string[] $documentation
   *  Documentation for the issue. Each key should be either a node ID on drupal.org
   *  or a full URL, and the values should be the link text (HTML allowed).
   */
  public function setIssue(ConverterInterface $converter, $issue_type, $title, $summary = NULL, $level = 'error', $documentation = []) {
    if (empty($this->issues[$issue_type])) {
      $this->issues[$issue_type] = [
        'identified_by' => $converter,
        'title' => $title,
        'summary' => $summary,
        'level' => $level,
        'documentation' => $documentation,
        'violations' => [],
      ];
    }
  }

  /**
   * Marks a violation of a particular issue.
   *
   * @param string $issue_type
   *  The issue type. This must already have been defined in a previous call to setIssue().
   * @param string $file
   *  The path to the file where the violation was found.
   * @param \Pharborist\SourcePosition $position
   *  Optional source position (line number and column) of the violation.
   *
   * @throws \DomainException if the issue type is not already defined.
   */
  public function setViolation($issue_type, $file, SourcePosition $position = NULL) {
    if (isset($this->issues[$issue_type])) {
      $issue = &$this->issues[$issue_type];

      if (! isset($issue['violations'][$file])) {
        $issue['violations'][$file] = [];
      }

      if ($position) {
        $issue['violations'][$file][] = $position;
      }
    }
    else {
      throw new \DomainException('Cannot set a violation of undefined issue type ' . $issue_type);
    }
  }

  public function generateReport() {
    $render = [
      '#theme' => 'dmu_report',
      '#issues' => $this->issues,
    ];
    $this->write('upgrade-info.html', render($render));
  }

  /**
   * Drops all open syntax trees.
   */
  public function releaseFiles() {
    $this->files = [];
  }

  /**
   * Parses a module file into a syntax tree and caches it in $this->files.
   *
   * @param string $path
   *  The file path.
   *
   * @return \Pharborist\TopNode
   */
  public function getFile($path) {
    if (! isset($this->files[$path])) {
      $this->files[$path] = Parser::parseFile($path);
    }
    return $this->files[$path];
  }

}
