<?php

/**
 * @file
 * Contains Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide\HookMenu.
 */

namespace Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide;

use Drupal\drupalmoduleupgrader\Converter\HookConverterBase;
use Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\Router as Drupal7Router;
use Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\RouteWrapper as Drupal7Route;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\LinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkIndex;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\LocalActionLinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\LocalTaskLinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\LinkBinding\MenuLinkBinding;
use Drupal\drupalmoduleupgrader\Converter\Routing\RouterBase as Drupal8Router;
use Drupal\drupalmoduleupgrader\Utility\ClassGeneratorTrait;

/**
 * @Converter(
 *  id = "HookMenu",
 *  description = @Translation("Converts Drupal 7's hook_menu() to shiny new routes."),
 *  documentation = {
 *    {
 *      "url" = "https://www.drupal.org/node/1800686",
 *      "title" = @Translation("All functionality of <code>hook_menu()</code> has been replaced")
 *    },
 *    {
 *      "url" = "https://www.drupal.org/node/2165243",
 *      "title" = @Translation("Contextual links are now plugins defined in <kbd>MODULE.links.contextual.yml</kbd>")
 *    },
 *    {
 *      "url" = "https://www.drupal.org/node/2007444",
 *      "title" = @Translation("Local actions are now plugins defined in <kbd>MODULE.links.action.yml</kbd>")
 *    },
 *    {
 *      "url" = "https://www.drupal.org/node/2044515",
 *      "title" = @Translation("Local tasks and now plugins defined in <kbd>MODULE.links.task.yml</kbd>")
 *    },
 *    {
 *      "url" = "https://www.drupal.org/node/2228089",
 *      "title" = @Translation("Menu links are now defined in <kbd>MODULE.links.menu.yml</kbd>")
 *    },
 *    {
 *      "url" = "https://www.drupal.org/developing/api/8/menu",
 *      "title" =  @Translation("Drupal 8 Menu API documentation")
 *    }
 *  },
 *  message = @Translation("hook_menu() has been replaced by a completely new routing and linking system."),
 *  hook = "menu",
 *  fixme = @Translation("hook_menu has been completely removed in Drupal 8.

Its functionality is now divided into routes (which map paths to callbacks) and links,
which are separate concepts. Routes are defined in MODULE.routing.yml (which has been
auto-generated for you), and links are defined in different YAML files, depending on
the type of link:

- Standard menu links are defined in MODULE.links.menu.yml.
- Local tasks (first level, second level, and defaults) are defined in MODULE.links.task.yml
- Local actions are defined in MODULE.links.action.yml
- Contextual links are defined in MODULE.links.contextual.yml.

If your implementation of hook_menu() includes any of those link types, an appropriate
YAML file will have been automatically generated for you.")
 * )
 */
class HookMenu extends HookConverterBase {

  use ClassGeneratorTrait;

  /**
   * Maps Drupal 7 route paths to Drupal 8 route names.
   *
   * @var array
   */
  private $routeMap = [];

  /**
   * The route converters' plugin manager.
   *
   * @var \Drupal\Component\Plugin\PluginManagerInterface
   */
  private $routeConverters;

  /**
   * @var \Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\Router
   */
  private $d7router;

  /**
   * @var \Drupal\drupalmoduleupgrader\Converter\Routing\RouterBase
   */
  private $d8router;

  /**
   * Constructs a HookMenu object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    $this->routeConverters = \Drupal::service('plugin.manager.drupalmoduleupgrader.route');
    $this->d7router = new Drupal7Router();
    $this->d8router = new Drupal8Router();
  }

  /**
   * Executes the original hook implementation and builds an index
   * of the Drupal 7 router items.
   *
   * This is a dumb kludge around a dumb bug (#2339357), and honestly
   * this entire thing should be re-freaking-factored so that this method
   * can be ditched.
   */
  private function init() {
    $items = $this->module->executeFunction($this->hook);
    foreach ($items as $path => $item) {
      $this->d7router->addRoute(new Drupal7Route($path, $item));
    }
    $this->d7router->finalize();
  }

  /**
   * {@inheritdoc}
   */
  public function analyze() {
    parent::analyze();

    $this->init();
    foreach ($this->d7router as $route) {
      if ($route->containsKey('page callback')) {
        $this->getRouteConverter($route)->analyze();
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function convert() {
    // Create the default controller.
    $controller = $this->generate();
    $this->module->addClass($controller);

    $this->init();
    foreach ($this->d7router as $path => $route) {
      if ($route->containsKey('page callback')) {
        $converter = $this->getRouteConverter($route);
        $d8_route = $converter->buildRoute();
        $converter->convert();
        $this->d8router->addRoute($d8_route);
        $this->routeMap[$path] = $d8_route->getIdentifier();
      }
    }
    $this->d8router->finalize();

    $routing = [];
    foreach ($this->d8router as $name => $route) {
      $routing[$name] = [
        'path' => $route->getPath()->__toString(),
        'defaults' => $route->getDefaults(),
        'requirements' => $route->getRequirements(),
      ];
    }
    $this->module->writeInfo('routing', $routing);

    // Default local tasks are really just aliases to their parent routes, so reflect
    // this in the route map.
    foreach ($this->d7router->getDefaultLocalTasks() as $path => $route) {
      $parent = $route->getPath()->getParent()->__toString();

      if (isset($this->routeMap[$parent])) {
        $this->routeMap[$path] = $this->routeMap[$parent];
      }
    }

    $links = [
      'menu' => new LinkIndex(),
      'task' => new LinkIndex(),
      'action' => new LinkIndex(),
      'contextual' => new LinkIndex(),
    ];

    foreach ($this->d7router->getAllLinks() as $path => $source) {
      $binding = LinkBinding::create($source, $this->d8router->get($this->routeMap[$path]));

      if ($binding instanceof MenuLinkBinding) {
        $links['menu']->addBinding($binding);
      }
      elseif ($binding instanceof LocalTaskLinkBinding) {
        $links['task']->addBinding($binding);
      }
      elseif ($binding instanceof LocalActionLinkBinding) {
        $links['action']->addBinding($binding);
      }
      elseif ($source->isContextualLink()) {
        $links['contextual']->addBinding($binding);
      }
    }

    $links = array_map(function(LinkIndex $index) {
      return $index->build();
    }, $links);

    foreach ($links['contextual'] as $link) {
      $link['group'] = $this->module->getMachineName();
    }

    foreach ($links as $group => $data) {
      if ($data) {
        $this->module->writeInfo('links.' . $group, $data);
      }
    }

    $this->addFixMe();
    $this->module->writeClass($controller);
    drush_log('Converted hook_menu() implementation.', 'success');
  }

  /**
   * Gets the appropriate route converter for a Drupal 7 router item, if one exists.
   *
   * @param \Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\Route $route
   *  The original Drupal 7 route.
   *
   * @return \Drupal\drupalmoduleupgrader\Converter\ConverterInterface
   */
  private function getRouteConverter(Drupal7Route $route) {
    $plugin_id = $route['page callback'];

    // Check if there is a route converter to handle this page callback.
    if (! $this->routeConverters->hasDefinition($plugin_id)) {
      $plugin_id = 'default';
    }

    return $this->routeConverters->createInstance($plugin_id, [
      'route' => $route,
      'module' => $this->module,
    ]);
  }

  /**
   * {@inheritdoc}
   */
  public function generate() {
    $render = [
      '#theme' => 'dmu_controller',
      '#module' => $this->module->getMachineName(),
    ];

    return $this->parse($render);
  }

}
