<?php

/**
 * @file
 * Contains Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide\InfoToYAML.
 */

namespace Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide;

use Drupal\drupalmoduleupgrader\Converter\ConverterBase;
use Symfony\Component\Yaml\Yaml;

/**
 * @Converter(
 *  id = "InfoToYAML",
 *  description = @Translation("Converts Drupal 7 info files to Drupal 8."),
 *  documentation = {
 *    {
 *      "url" = "https://www.drupal.org/node/1935708",
 *      "title" = @Translation("<kbd>.info</kbd> files are now <kbd>.info.yml</kbd> files")
 *    }
 *  },
 *  message = @Translation("<kbd>.info</kbd> files are now <kbd>.info.yml</kbd> files.")
 * )
 */
class InfoToYAML extends ConverterBase {

  /**
   * The parsed D7 info file.
   *
   * @var array
   */
  private $info;

  /**
   * Constructs an InfoToYAML object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->info = self::parseInfo($this->module->getPath('.info'));
  }

  /**
   * {@inheritdoc}
   */
  public function analyze() {
    $documentation = $this->pluginDefinition['documentation'];

    $this->module->setIssue($this, 'InfoToYAML_CoreKey', 'Module info files must contain a <var>core</var> key');
    $this->module->setIssue($this, 'InfoToYAML_TypeKey', 'Module info file must contain a <var>type</var> key');

    if (isset($this->info['dependencies'])) {
      $this->module->setIssue($this, 'InfoToYAML_Dependencies', 'Common dependencies have moved into core', NULL, 'warning', $documentation);
    }
    if (isset($this->info['files'])) {
      $this->module->setIssue($this, 'InfoToYAML_Files', 'Modules do not need to declare files in their info file', NULL, 'warning', $documentation);
    }
    if (isset($this->info['configure'])) {
      $this->module->setIssue($this, 'InfoToYAML_Configure', "Module info files' <var>configure</var> key must be a route name", NULL, 'error', $documentation);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function convert() {
    $this->info['core'] = '8.x';
    $this->info['type'] = 'module';

    if (isset($this->info['dependencies'])) {
      // array_values() is called in order to reindex the array. Issue #2340207
      $this->info['dependencies'] = array_values(array_diff($this->info['dependencies'], ['ctools', 'list']));
    }

    unset($this->info['files'], $this->info['configure']);

    $this->module->writeInfo('info', $this->info);
    drush_log('Converted legacy info file to YAML.', 'success');
  }

  /**
   * Parses a D7 info file. This is copied straight outta the D7 function 
   * drupal_parse_info_format().
   */
  public static function parseInfo($file) {
    $info = [];
    $constants = get_defined_constants();
    $data = file_get_contents($file);

    if (preg_match_all('
      @^\s*                           # Start at the beginning of a line, ignoring leading whitespace
      ((?:
        [^=;\[\]]|                    # Key names cannot contain equal signs, semi-colons or square brackets,
        \[[^\[\]]*\]                  # unless they are balanced and not nested
      )+?)
      \s*=\s*                         # Key/value pairs are separated by equal signs (ignoring white-space)
      (?:
        ("(?:[^"]|(?<=\\\\)")*")|     # Double-quoted string, which may contain slash-escaped quotes/slashes
        (\'(?:[^\']|(?<=\\\\)\')*\')| # Single-quoted string, which may contain slash-escaped quotes/slashes
        ([^\r\n]*?)                   # Non-quoted string
      )\s*$                           # Stop at the next end of a line, ignoring trailing whitespace
      @msx', $data, $matches, PREG_SET_ORDER)) {
      foreach ($matches as $match) {
        // Fetch the key and value string.
        $i = 0;
        foreach (array('key', 'value1', 'value2', 'value3') as $var) {
          $$var = isset($match[++$i]) ? $match[$i] : '';
        }
        $value = stripslashes(substr($value1, 1, -1)) . stripslashes(substr($value2, 1, -1)) . $value3;

        // Parse array syntax.
        $keys = preg_split('/\]?\[/', rtrim($key, ']'));
        $last = array_pop($keys);
        $parent = &$info;

        // Create nested arrays.
        foreach ($keys as $key) {
          if ($key == '') {
            $key = count($parent);
          }
          if (!isset($parent[$key]) || !is_array($parent[$key])) {
            $parent[$key] = array();
          }
          $parent = &$parent[$key];
        }

        // Handle PHP constants.
        if (isset($constants[$value])) {
          $value = $constants[$value];
        }

        // Insert actual value.
        if ($last == '') {
          $last = count($parent);
        }
        $parent[$last] = $value;
      }
    }

    return $info;
  }

}
