<?php

/**
 * @file
 * Contains Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide\PSR4.
 */

namespace Drupal\drupalmoduleupgrader\Plugin\DMU\ModuleWide;

use Drupal\drupalmoduleupgrader\Converter\ConverterBase;
use Drupal\drupalmoduleupgrader\ModuleContext;
use Pharborist\ClassNode;
use Pharborist\Filter;
use Pharborist\NodeCollection;
use Pharborist\TopNode;
use Pharborist\WhitespaceNode;

/**
 * @Converter(
 *  id = "PSR4",
 *  description = @Translation("Moves classes into PSR-4 directory structure."),
 *  documentation = {
 *    {
 *      "url" = "https://www.drupal.org/node/2246699",
 *      "title" = @Translation("PSR-4 compatible class loader added to Drupal core")
 *    }
 *  },
 *  message = @Translation("Classes must be PSR-4 compliant.")
 * )
 */
class PSR4 extends ConverterBase {

  /**
   * {@inheritdoc}
   */
  public function convert() {
    $targets = new NodeCollection([]);

    foreach ($this->module->getObjectNames('class') as $class) {
      $classes = $this->module
        ->getFileContaining('class', $class)
        ->children(Filter::isInstanceOf('\Pharborist\ClassNode'))
        ->not(function(ClassNode $class) {
          // Don't handle tests -- we've got a separate plugin for that.
          // We need to use the (string) cast on getExtends() because it may return
          // NULL if the class is not extending another.
          return in_array((string) $class->getExtends(), [ 'DrupalWebTestCase', 'DrupalUnitTestCase' ]);
        });

      $targets->add($classes);
    }

    $targets->each(function(ClassNode $class) {
      $this->module->writeClass(self::toPSR4($this->module, $class));
    });
  }

  /**
   * Utility method to PSR4-ify a class. It'll move the class into its own file
   * in the given module's namespace. The class is modified in-place, so you
   * should clone it before calling this function if you want to make a PSR-4
   * *copy* of it.
   *
   * @param \Drupal\drupalmoduleupgrader\ModuleContext $module
   *  The module which will own the class.
   * @param \Pharborist\ClassNode $class
   *  The class to modify.
   *
   * @return \Pharborist\ClassNode
   *  The modified class, returned for convenience.
   */
  public static function toPSR4(ModuleContext $module, ClassNode $class) {
    $ns = $module->getNS();
    TopNode::create($ns)->getNamespace($ns)->append($class->remove());
    WhitespaceNode::create("\n\n")->insertBefore($class);

    return $class;
  }

}
