<?php

/**
 * @file
 * Contains Drupal\drupalmoduleupgrader\Plugin\DMU\Routing\FormRoute.
 */

namespace Drupal\drupalmoduleupgrader\Plugin\DMU\Routing;

use Drupal\drupalmoduleupgrader\Converter\Routing\Drupal7\ParameterMap;
use Drupal\drupalmoduleupgrader\Converter\Routing\RouteConverterBase;
use Drupal\drupalmoduleupgrader\Utility\ConfigFormConverter;
use Drupal\drupalmoduleupgrader\Utility\FormConverter;
use Pharborist\Filter;

/**
 * @Converter(
 *  id = "drupal_get_form",
 *  description = @Translation("Converts a drupal_get_form() menu item to a _form route."),
 *  message = @Translation("drupal_get_form() is no longer a valid page callback and must be replaced with a route to a class implementing \Drupal\Core\Form\FormInterface."),
 * )
 */
class FormRoute extends RouteConverterBase {

  /**
   * The original form ID.
   *
   * @var string
   */
  private $formID;

  /**
   * Constructs a FormRoute object.
   */
  public function __construct(array $configuration, $plugin_id, $plugin_definition) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);

    // Get the form ID. If drupal_get_form() was the page callback, the form ID
    // *will* be the first argument, or this is a malformed route that deserves
    // every error it causes.
    $form_builder = $this->route['page arguments'][0];

    // If the module doesn't define the form builder, getFunction() will throw an exception.
    // But then again, if the module doesn't define the form builder...we can't really
    // do anything to convert the route, now can we?
    $this->route['page callback'] = clone $this->module->getFunction($form_builder);
    $this->formID = $form_builder;

    // Ignore the form builder's first two parameters, which will always be $form and $form_state;
    // they are never part of the route.
    $parameters = array_slice($this->route['page callback']->getParameters(), 2);

    // Shift the form ID off the arguments array.
    $arguments = array_slice($this->route['page arguments'], 1);

    // Both arrays need to start at the same index in order for ParameterMap to work
    // properly. ParameterMap expects the calling code to do this step.
    $parameters = $this->bumpKeys($parameters, 2);
    $arguments = $this->bumpKeys($arguments, 2);

    $this->route['page arguments'] = new ParameterMap($this->route->getPath(), $parameters, $arguments);

    $module_name = $this->module->getMachineName();
    $this->name = $module_name . '.' . $this->unPrefix($this->formID, $module_name);
    $arguments = array_filter($arguments, 'is_string');
    if ($arguments) {
      $this->name .= '_' . implode('_', $arguments);
    }
  }

  /**
   * Returns a copy of the input array with the keys increased by $offset. This
   * only works on numerically indexed arrays; I don't know what it does to
   * associative arrays, but probably nothing good.
   *
   * @param array $input
   *  The input array.
   *
   * @param int $offset
   *  The offset to add to the keys.
   *
   * @return array
   */
  private function bumpKeys(array $input, $offset = 0) {
    $output = [];

    foreach ($input as $key => $value) {
      $output[ $key + $offset ] = $value;
    }

    return $output;
  }

  /**
   * Returns an instance of the appropriate form converter. At the time of
   * this writing, that'll be the default converter, unless the form builder
   * calls system_settings_form(), in which case it'll be the config form
   * converter.
   *
   * @return \Drupal\Component\Plugin\PluginInterface
   */
  private function getFormConverter() {
    $builder = $this->route['page callback'];

    if ($builder->getBody()->has(Filter::isFunctionCall('system_settings_form'))) {
      return new ConfigFormConverter($this->module, $builder);
    }
    else {
      return new FormConverter($this->module, $builder);
    }
  }

  /**
   * {@inheritdoc}
   */
  public function convert() {
    // Convert the form to a controller class, if it hasn't already been done.
    if (! $this->module->definesForm($this->formID)) {
      $this->getFormConverter()->convert();
      drush_log("Created a form controller for {$this->formID}.", 'success');
    }
    $this->controller = $this->module->getForm($this->formID);

    parent::convert();
  }

  /**
   * {@inheritdoc}
   */
  public function buildRoute() {
    $route = parent::buildRoute();
    $route->setDefault('_form', $this->module->getNS() . '\\Form\\' . $this->toTitleCase($this->formID));

    return $route;
  }

}
