<?php

/**
 * @file
 * Contains \Drupal\drupalmoduleupgrader\Utility\FormConverter.
 */

namespace Drupal\drupalmoduleupgrader\Utility;

use Drupal\drupalmoduleupgrader\ModuleContext;
use Pharborist\Filter;
use Pharborist\FunctionDeclarationNode;
use Pharborist\ParameterNode;
use Pharborist\ClassNode;
use Pharborist\Parser;
use Pharborist\TokenNode;

/**
 * Converts a form from a set of callback functions to a class implementing
 * \Drupal\Core\Form\FormInterface.
 */
class FormConverter {

  use ClassGeneratorTrait, StringTransformTrait;

  /**
   * @var \Drupal\drupalmoduleupgrader\ModuleContext
   */
  protected $module;

  /**
   * The form builder function.
   *
   * @var \Pharborist\FunctionDeclarationNode
   */
  protected $builder;

  /**
   * Constructs a FormConverter object.
   */
  public function __construct(ModuleContext $module, FunctionDeclarationNode $builder) {
    $this->module = $module;
    $this->builder = $builder;
  }

  /**
   * {@inheritdoc}
   */
  public function convert() {
    $this->module->forms[ $this->builder->getName()->__toString() ] = $this->generate();
  }

  /**
   * Generates a renderable array for the form controller.
   *
   * @return array
   */
  protected function getRenderable() {
    $render = [
      '#theme' => 'dmu_form',
      '#module' => $this->module->getMachineName(),
      '#form_id' => $this->builder->getName()->__toString(),
    ];
    $render['#class'] = $this->toTitleCase($render['#form_id']);

    return $render;
  }

  /**
   * {@inheritdoc}
   */
  public function generate() {
    $controller = Parser::parseSource(render($this->getRenderable()))
      ->find(Filter::isInstanceOf('\Pharborist\ClassNode'))
      ->get(0);

    $this->addMethod($this->builder, $controller, 'buildForm');

    $validator = $this->getHandler('validate');
    if ($validator) {
      $this->addMethod($validator, $controller);
    }

    $submitter = $this->getHandler('submit');
    if ($submitter) {
      $this->addMethod($submitter, $controller);
    }

    return $controller;
  }

  /**
   * Tries to retrieve the form's default validate or submit handler.
   *
   * @param string $type
   *  The handler type (validate or submit).
   *
   * @return null|FunctionDeclarationNode
   */
  protected function getHandler($type) {
    $handler = $this->builder->getName() . '_' . $type;

    if ($this->module->defines('function', $handler)) {
      /** @var \Pharborist\FunctionDeclarationNode $function */
      $function = $this->module->getFunction($handler);
      // The first argument ($form) is always passed by reference.
      $function->getArguments()[0]->setReference(TRUE);

      switch ($type) {
        case 'validate':
          $function->setName('validateForm');
          break;
        case 'submit':
          $function->setName('submitForm');
          break;
        default:
          break;
      }

      return $function;
    }
  }

  /**
   * {@inheritdoc}
   */
  protected function addMethod(FunctionDeclarationNode $function, ClassNode $class, $alias = NULL) {
    // Detach the function from its syntax tree so it can be moved into the
    // destination class. If we fail to do this, things can go very wrong.
    // @see \Drupal\drupalmoduleupgrader\Converter\Routing\RouteConverterBase::transferCallback().
    $function->remove();

    // Add the parameters required for FormInterface conformance.
    $parameters = $function->getParameters();
    if (empty($parameters)) {
      $parameters[0] = ParameterNode::create('$form');
      $function->appendParameter($parameters[0]);
    }
    if (sizeof($parameters) == 1) {
      $parameters[1] = ParameterNode::create('$form_state');
      $function->appendParameter($parameters[1]);
    }

    // Provide type hints, again in conformance with FormInterface.
    $parameters[0]->setTypeHint('array');
    $parameters[1]->setTypeHint('FormStateInterface');

    // Additional parameters MUST have a default value of NULL in order to conform
    // to FormInterface.
    for ($i = 2; $i < sizeof($parameters); $i++) {
      $parameters[$i]->setValue(new TokenNode(T_STRING, 'NULL'));
    }

    // @todo Get $form_state's local name and use it for the conversions below.

    // @todo
    // Rewrite calls to form_set_error(), form_clear_error(), form_get_errors(),
    // form_get_error(), form_error(), and form_set_value() to their $form_state
    // method counterparts. That's the annoying thing about devising a stable
    // API: the next step is busywork. None of these rewrites can be done by
    // the Grep plugin because the name of the $form_state variable may vary.
    // @see https://www.drupal.org/node/2121003

    // Convert form_set_error() to $form_state->setErrorByName()
    //  D7: form_set_error($name = NULL, $message = '', $limit_validation_errors = NULL)
    //  D8: setErrorByName($name, $message = '')

    // Convert form_clear_error() to $form_state->clearErrors()
    //  D7: form_clear_error()
    //  D8: clearErrors()

    // Convert form_get_errors() to $form_state->getErrors()
    //  D7: form_get_errors()
    //  D8: getErrors()

    // Convert form_get_error() to $form_state->getError()
    //  D7: form_get_error($element)
    //  D8: getError($element)

    // Convert form_set_value() to $form_state->setValueForElement()
    // NOTE: This one is *really* tricky because $key is basically $element['#parents']
    //  D7: form_set_value($element, $value, &$form_state)
    //  D8: setValue($key, $value)

    $function->cloneAsMethodOf($class)->setName($alias);
  }

}
